# -*- coding: utf-8 -*-
"""
Unit tests for the spike_train_correlation module.

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import math
import unittest

import neo
from neo.io import NixIO
import numpy as np
import quantities as pq
from numpy.testing import assert_array_equal, assert_array_almost_equal

import elephant.conversion as conv
import elephant.spike_train_correlation as sc
from elephant.spike_train_generation import StationaryPoissonProcess, \
    StationaryGammaProcess
import math
from elephant.datasets import download_datasets
from elephant.spike_train_generation import homogeneous_poisson_process, \
    homogeneous_gamma_process


class CovarianceTestCase(unittest.TestCase):

    def setUp(self):
        # These two arrays must be such that they do not have coincidences
        # spanning across two neighbor bins assuming ms bins [0,1),[1,2),...
        self.test_array_1d_0 = [
            1.3, 7.56, 15.87, 28.23, 30.9, 34.2, 38.2, 43.2]
        self.test_array_1d_1 = [
            1.02, 2.71, 18.82, 28.46, 28.79, 43.6]

        # Build spike trains
        self.st_0 = neo.SpikeTrain(
            self.test_array_1d_0, units='ms', t_stop=50.)
        self.st_1 = neo.SpikeTrain(
            self.test_array_1d_1, units='ms', t_stop=50.)

        # And binned counterparts
        self.binned_st = conv.BinnedSpikeTrain(
            [self.st_0, self.st_1], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)

    def test_covariance_binned(self):
        """
        Test covariance between two binned spike trains.
        """

        # Calculate clipped and unclipped
        res_clipped = sc.covariance(
            self.binned_st, binary=True, fast=False)
        res_unclipped = sc.covariance(
            self.binned_st, binary=False, fast=False)

        # Check dimensions
        self.assertEqual(len(res_clipped), 2)
        self.assertEqual(len(res_unclipped), 2)

        # Check result unclipped against result calculated from scratch for
        # the off-diagonal element
        mat = self.binned_st.to_array()
        mean_0 = np.mean(mat[0])
        mean_1 = np.mean(mat[1])
        target_from_scratch = \
            np.dot(mat[0] - mean_0, mat[1] - mean_1) / (len(mat[0]) - 1)

        # Check result unclipped against result calculated by numpy.corrcoef
        target_numpy = np.cov(mat)

        self.assertAlmostEqual(target_from_scratch, target_numpy[0][1])
        self.assertAlmostEqual(res_unclipped[0][1], target_from_scratch)
        self.assertAlmostEqual(res_unclipped[1][0], target_from_scratch)

        # Check result clipped against result calculated from scratch for
        # the off-diagonal elemant
        mat = self.binned_st.to_bool_array()
        mean_0 = np.mean(mat[0])
        mean_1 = np.mean(mat[1])
        target_from_scratch = \
            np.dot(mat[0] - mean_0, mat[1] - mean_1) / (len(mat[0]) - 1)

        # Check result unclipped against result calculated by numpy.corrcoef
        target_numpy = np.cov(mat)

        self.assertAlmostEqual(target_from_scratch, target_numpy[0][1])
        self.assertAlmostEqual(res_clipped[0][1], target_from_scratch)
        self.assertAlmostEqual(res_clipped[1][0], target_from_scratch)

    def test_covariance_binned_same_spiketrains(self):
        """
        Test if the covariation between two identical binned spike
        trains evaluates to the expected 2x2 matrix.
        """
        # Calculate correlation
        binned_st = conv.BinnedSpikeTrain(
            [self.st_0, self.st_0], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)
        result = sc.covariance(binned_st, fast=False)

        # Check dimensions
        self.assertEqual(len(result), 2)
        # Check result
        assert_array_equal(result[0][0], result[1][1])

    def test_covariance_binned_short_input(self):
        """
        Test if input list of only one binned spike train yields correct result
        that matches numpy.cov (covariance with itself)
        """
        # Calculate correlation
        binned_st = conv.BinnedSpikeTrain(
            self.st_0, t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)
        result = sc.covariance(binned_st, binary=True, fast=False)

        # Check result unclipped against result calculated by numpy.corrcoef
        mat = binned_st.to_bool_array()
        target = np.cov(mat)

        # Check result and dimensionality of result
        self.assertEqual(result.ndim, target.ndim)
        assert_array_almost_equal(result, target)
        assert_array_almost_equal(target,
                                  sc.covariance(binned_st, binary=True,
                                                fast=True))

    def test_covariance_fast_mode(self):
        np.random.seed(27)
        st = StationaryPoissonProcess(rate=10 * pq.Hz, t_stop=10 * pq.s
                                      ).generate_spiketrain()
        binned_st = conv.BinnedSpikeTrain(st, n_bins=10)
        assert_array_almost_equal(sc.covariance(binned_st, fast=False),
                                  sc.covariance(binned_st, fast=True))


class CorrCoefTestCase(unittest.TestCase):

    def setUp(self):
        # These two arrays must be such that they do not have coincidences
        # spanning across two neighbor bins assuming ms bins [0,1),[1,2),...
        self.test_array_1d_0 = [
            1.3, 7.56, 15.87, 28.23, 30.9, 34.2, 38.2, 43.2]
        self.test_array_1d_1 = [
            1.02, 2.71, 18.82, 28.46, 28.79, 43.6]
        self.test_array_1d_2 = []

        # Build spike trains
        self.st_0 = neo.SpikeTrain(
            self.test_array_1d_0, units='ms', t_stop=50.)
        self.st_1 = neo.SpikeTrain(
            self.test_array_1d_1, units='ms', t_stop=50.)
        self.st_2 = neo.SpikeTrain(
            self.test_array_1d_2, units='ms', t_stop=50.)

        # And binned counterparts
        self.binned_st = conv.BinnedSpikeTrain(
            [self.st_0, self.st_1], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)

    def test_corrcoef_binned(self):
        """
        Test the correlation coefficient between two binned spike trains.
        """

        # Calculate clipped and unclipped
        res_clipped = sc.correlation_coefficient(
            self.binned_st, binary=True)
        res_unclipped = sc.correlation_coefficient(
            self.binned_st, binary=False)

        # Check dimensions
        self.assertEqual(len(res_clipped), 2)
        self.assertEqual(len(res_unclipped), 2)

        # Check result unclipped against result calculated from scratch for
        # the off-diagonal element
        mat = self.binned_st.to_array()
        mean_0 = np.mean(mat[0])
        mean_1 = np.mean(mat[1])
        target_from_scratch = \
            np.dot(mat[0] - mean_0, mat[1] - mean_1) / \
            np.sqrt(
                np.dot(mat[0] - mean_0, mat[0] - mean_0) *
                np.dot(mat[1] - mean_1, mat[1] - mean_1))

        # Check result unclipped against result calculated by numpy.corrcoef
        target_numpy = np.corrcoef(mat)

        self.assertAlmostEqual(target_from_scratch, target_numpy[0][1])
        self.assertAlmostEqual(res_unclipped[0][1], target_from_scratch)
        self.assertAlmostEqual(res_unclipped[1][0], target_from_scratch)

        # Check result clipped against result calculated from scratch for
        # the off-diagonal elemant
        mat = self.binned_st.to_bool_array()
        mean_0 = np.mean(mat[0])
        mean_1 = np.mean(mat[1])
        target_from_scratch = \
            np.dot(mat[0] - mean_0, mat[1] - mean_1) / \
            np.sqrt(
                np.dot(mat[0] - mean_0, mat[0] - mean_0) *
                np.dot(mat[1] - mean_1, mat[1] - mean_1))

        # Check result unclipped against result calculated by numpy.corrcoef
        target_numpy = np.corrcoef(mat)

        self.assertAlmostEqual(target_from_scratch, target_numpy[0][1])
        self.assertAlmostEqual(res_clipped[0][1], target_from_scratch)
        self.assertAlmostEqual(res_clipped[1][0], target_from_scratch)

    def test_corrcoef_binned_same_spiketrains(self):
        """
        Test if the correlation coefficient between two identical binned spike
        trains evaluates to a 2x2 matrix of ones.
        """
        # Calculate correlation
        binned_st = conv.BinnedSpikeTrain(
            [self.st_0, self.st_0], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)
        result = sc.correlation_coefficient(binned_st, fast=False)
        target = np.ones((2, 2))

        # Check dimensions
        self.assertEqual(len(result), 2)
        # Check result
        assert_array_almost_equal(result, target)
        assert_array_almost_equal(
            result, sc.correlation_coefficient(
                binned_st, fast=True))

    def test_corrcoef_binned_short_input(self):
        """
        Test if input list of one binned spike train yields 1.0.
        """
        # Calculate correlation
        binned_st = conv.BinnedSpikeTrain(
            self.st_0, t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)
        result = sc.correlation_coefficient(binned_st, fast=False)
        target = np.array(1.)

        # Check result and dimensionality of result
        self.assertEqual(result.ndim, 0)
        assert_array_almost_equal(result, target)
        assert_array_almost_equal(
            result, sc.correlation_coefficient(
                binned_st, fast=True))

    def test_empty_spike_train(self):
        """
        Test whether a warning is yielded in case of empty spike train.
        Also check correctness of the output array.
        """
        # st_2 is empty
        binned_12 = conv.BinnedSpikeTrain([self.st_1, self.st_2],
                                          bin_size=1 * pq.ms)

        with self.assertWarns(UserWarning):
            result = sc.correlation_coefficient(binned_12, fast=False)

        # test for NaNs in the output array
        target = np.zeros((2, 2)) * np.NaN
        target[0, 0] = 1.0
        assert_array_almost_equal(result, target)

    def test_corrcoef_fast_mode(self):
        np.random.seed(27)
        st = StationaryPoissonProcess(rate=10 * pq.Hz, t_stop=10 * pq.s
                                      ).generate_spiketrain()
        binned_st = conv.BinnedSpikeTrain(st, n_bins=10)
        assert_array_almost_equal(
            sc.correlation_coefficient(
                binned_st, fast=False), sc.correlation_coefficient(
                binned_st, fast=True))


class CrossCorrelationHistogramTest(unittest.TestCase):

    def setUp(self):
        # These two arrays must be such that they do not have coincidences
        # spanning across two neighbor bins assuming ms bins [0,1),[1,2),...
        self.test_array_1d_1 = [
            1.3, 7.56, 15.87, 28.23, 30.9, 34.2, 38.2, 43.2]
        self.test_array_1d_2 = [
            1.02, 2.71, 18.82, 28.46, 28.79, 43.6]

        # Build spike trains
        self.st_1 = neo.SpikeTrain(
            self.test_array_1d_1, units='ms', t_stop=50.)
        self.st_2 = neo.SpikeTrain(
            self.test_array_1d_2, units='ms', t_stop=50.)

        # And binned counterparts
        self.binned_st1 = conv.BinnedSpikeTrain(
            [self.st_1], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)
        self.binned_st2 = conv.BinnedSpikeTrain(
            [self.st_2], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)
        self.binned_sts = conv.BinnedSpikeTrain(
            [self.st_1, self.st_2], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)

        # Binned sts to check errors raising
        self.st_check_bin_size = conv.BinnedSpikeTrain(
            [self.st_1], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=5 * pq.ms)
        self.st_check_t_start = conv.BinnedSpikeTrain(
            [self.st_1], t_start=1 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)
        self.st_check_t_stop = conv.BinnedSpikeTrain(
            [self.st_1], t_start=0 * pq.ms, t_stop=40. * pq.ms,
            bin_size=1 * pq.ms)
        self.st_check_dimension = conv.BinnedSpikeTrain(
            [self.st_1, self.st_2], t_start=0 * pq.ms, t_stop=50. * pq.ms,
            bin_size=1 * pq.ms)

    def test_cross_correlation_histogram(self):
        """
        Test generic result of a cross-correlation histogram between two binned
        spike trains.
        """
        # Calculate CCH using Elephant (normal and binary version) with
        # mode equal to 'full' (whole spike trains are correlated)
        cch_clipped, bin_ids_clipped = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='full',
            binary=True)
        cch_unclipped, bin_ids_unclipped = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='full', binary=False)

        cch_clipped_mem, bin_ids_clipped_mem = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='full',
            binary=True, method='memory')
        cch_unclipped_mem, bin_ids_unclipped_mem = \
            sc.cross_correlation_histogram(
                self.binned_st1, self.binned_st2, window='full',
                binary=False, method='memory')
        # Check consistency two methods
        assert_array_equal(
            np.squeeze(cch_clipped.magnitude), np.squeeze(
                cch_clipped_mem.magnitude))
        assert_array_equal(
            np.squeeze(cch_clipped.times), np.squeeze(
                cch_clipped_mem.times))
        assert_array_equal(
            np.squeeze(cch_unclipped.magnitude), np.squeeze(
                cch_unclipped_mem.magnitude))
        assert_array_equal(
            np.squeeze(cch_unclipped.times), np.squeeze(
                cch_unclipped_mem.times))
        assert_array_almost_equal(bin_ids_clipped, bin_ids_clipped_mem)
        assert_array_almost_equal(bin_ids_unclipped, bin_ids_unclipped_mem)

        # Check normal correlation Note: Use numpy correlate to verify result.
        # Note: numpy conventions for input array 1 and input array 2 are
        # swapped compared to Elephant!
        mat1 = self.binned_st1.to_array()[0]
        mat2 = self.binned_st2.to_array()[0]
        target_numpy = np.correlate(mat2, mat1, mode='full')
        assert_array_equal(
            target_numpy, np.squeeze(cch_unclipped.magnitude))

        # Check cross correlation function for several displacements tau
        # Note: Use Elephant corrcoeff to verify result
        tau = [-25.0, 0.0, 13.0]  # in ms
        for t in tau:
            # adjust t_start, t_stop to shift by tau
            t0 = np.min([self.st_1.t_start + t * pq.ms, self.st_2.t_start])
            t1 = np.max([self.st_1.t_stop + t * pq.ms, self.st_2.t_stop])
            st1 = neo.SpikeTrain(self.st_1.magnitude + t, units='ms',
                                 t_start=t0 * pq.ms, t_stop=t1 * pq.ms)
            st2 = neo.SpikeTrain(self.st_2.magnitude, units='ms',
                                 t_start=t0 * pq.ms, t_stop=t1 * pq.ms)
            binned_sts = conv.BinnedSpikeTrain([st1, st2],
                                               bin_size=1 * pq.ms,
                                               t_start=t0 * pq.ms,
                                               t_stop=t1 * pq.ms)
            # caluclate corrcoef
            corrcoef = sc.correlation_coefficient(binned_sts)[1, 0]

            # expand t_stop to have two spike trains with same length as st1,
            # st2
            st1 = neo.SpikeTrain(self.st_1.magnitude, units='ms',
                                 t_start=self.st_1.t_start,
                                 t_stop=self.st_1.t_stop + np.abs(t) * pq.ms)
            st2 = neo.SpikeTrain(self.st_2.magnitude, units='ms',
                                 t_start=self.st_2.t_start,
                                 t_stop=self.st_2.t_stop + np.abs(t) * pq.ms)
            binned_st1 = conv.BinnedSpikeTrain(
                st1, t_start=0 * pq.ms, t_stop=(50 + np.abs(t)) * pq.ms,
                bin_size=1 * pq.ms)
            binned_st2 = conv.BinnedSpikeTrain(
                st2, t_start=0 * pq.ms, t_stop=(50 + np.abs(t)) * pq.ms,
                bin_size=1 * pq.ms)
            # calculate CCHcoef and take value at t=tau
            CCHcoef, _ = sc.cch(binned_st1, binned_st2,
                                cross_correlation_coefficient=True)
            left_edge = - binned_st1.n_bins + 1
            tau_bin = int(t / float(binned_st1.bin_size.magnitude))
            assert_array_almost_equal(
                corrcoef, CCHcoef[tau_bin - left_edge].magnitude)

        # Check correlation using binary spike trains
        mat1 = np.array(self.binned_st1.to_bool_array()[0], dtype=int)
        mat2 = np.array(self.binned_st2.to_bool_array()[0], dtype=int)
        target_numpy = np.correlate(mat2, mat1, mode='full')
        assert_array_equal(
            target_numpy, np.squeeze(cch_clipped.magnitude))

        # Check the time axis and bin IDs of the resulting AnalogSignal
        assert_array_almost_equal(
            (bin_ids_clipped - 0.5) * self.binned_st1.bin_size,
            cch_unclipped.times)
        assert_array_almost_equal(
            (bin_ids_clipped - 0.5) * self.binned_st1.bin_size,
            cch_clipped.times)

        # Calculate CCH using Elephant (normal and binary version) with
        # mode equal to 'valid' (only completely overlapping intervals of the
        # spike trains are correlated)
        cch_clipped, bin_ids_clipped = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='valid',
            binary=True)
        cch_unclipped, bin_ids_unclipped = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='valid',
            binary=False)
        cch_clipped_mem, bin_ids_clipped_mem = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='valid',
            binary=True, method='memory')
        cch_unclipped_mem, bin_ids_unclipped_mem = \
            sc.cross_correlation_histogram(
                self.binned_st1, self.binned_st2, window='valid',
                binary=False, method='memory')

        # Check consistency two methods
        assert_array_equal(
            np.squeeze(cch_clipped.magnitude), np.squeeze(
                cch_clipped_mem.magnitude))
        assert_array_equal(
            np.squeeze(cch_clipped.times), np.squeeze(
                cch_clipped_mem.times))
        assert_array_equal(
            np.squeeze(cch_unclipped.magnitude), np.squeeze(
                cch_unclipped_mem.magnitude))
        assert_array_equal(
            np.squeeze(cch_unclipped.times), np.squeeze(
                cch_unclipped_mem.times))
        assert_array_equal(bin_ids_clipped, bin_ids_clipped_mem)
        assert_array_equal(bin_ids_unclipped, bin_ids_unclipped_mem)

        # Check normal correlation Note: Use numpy correlate to verify result.
        # Note: numpy conventions for input array 1 and input array 2 are
        # swapped compared to Elephant!
        mat1 = self.binned_st1.to_array()[0]
        mat2 = self.binned_st2.to_array()[0]
        target_numpy = np.correlate(mat2, mat1, mode='valid')
        assert_array_equal(
            target_numpy, np.squeeze(cch_unclipped.magnitude))

        # Check correlation using binary spike trains
        mat1 = np.array(self.binned_st1.to_bool_array()[0], dtype=int)
        mat2 = np.array(self.binned_st2.to_bool_array()[0], dtype=int)
        target_numpy = np.correlate(mat2, mat1, mode='valid')
        assert_array_equal(
            target_numpy, np.squeeze(cch_clipped.magnitude))

        # Check the time axis and bin IDs of the resulting AnalogSignal
        assert_array_equal(
            (bin_ids_clipped - 0.5) * self.binned_st1.bin_size,
            cch_unclipped.times)
        assert_array_equal(
            (bin_ids_clipped - 0.5) * self.binned_st1.bin_size,
            cch_clipped.times)

        # Check for wrong window parameter setting
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram, self.binned_st1,
            self.binned_st2, window='dsaij')
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram, self.binned_st1,
            self.binned_st2, window='dsaij', method='memory')

    def test_raising_error_wrong_inputs(self):
        """Check that an exception is thrown if the two spike trains are not
        fullfilling the requirement of the function"""
        # Check the bin_sizes are the same
        self.assertRaises(
            ValueError,
            sc.cross_correlation_histogram, self.binned_st1,
            self.st_check_bin_size)
        # Check input are one dimensional
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram,
            self.st_check_dimension, self.binned_st2)
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram,
            self.binned_st2, self.st_check_dimension)

    def test_window(self):
        """Test if the window parameter is correctly interpreted."""
        cch_win, bin_ids = sc.cch(
            self.binned_st1, self.binned_st2, window=[-30, 30])
        cch_win_mem, bin_ids_mem = sc.cch(
            self.binned_st1, self.binned_st2, window=[-30, 30],
            method='memory')

        self.assertEqual(len(bin_ids), cch_win.shape[0])
        assert_array_equal(bin_ids, np.arange(-30, 31, 1))
        assert_array_equal(
            (bin_ids - 0.5) * self.binned_st1.bin_size, cch_win.times)

        assert_array_equal(bin_ids_mem, np.arange(-30, 31, 1))
        assert_array_equal(
            (bin_ids_mem - 0.5) * self.binned_st1.bin_size, cch_win.times)

        assert_array_equal(cch_win, cch_win_mem)
        cch_unclipped, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='full', binary=False)
        assert_array_equal(cch_win, cch_unclipped[19:80])

        _, bin_ids = sc.cch(
            self.binned_st1, self.binned_st2, window=[20, 30])
        _, bin_ids_mem = sc.cch(
            self.binned_st1, self.binned_st2, window=[20, 30], method='memory')

        assert_array_equal(bin_ids, np.arange(20, 31, 1))
        assert_array_equal(bin_ids_mem, np.arange(20, 31, 1))

        _, bin_ids = sc.cch(
            self.binned_st1, self.binned_st2, window=[-30, -20])

        _, bin_ids_mem = sc.cch(
            self.binned_st1, self.binned_st2, window=[-30, -20],
            method='memory')

        assert_array_equal(bin_ids, np.arange(-30, -19, 1))
        assert_array_equal(bin_ids_mem, np.arange(-30, -19, 1))

        # Check for wrong assignments to the window parameter
        # Test for window longer than the total length of the spike trains
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram, self.binned_st1,
            self.binned_st2, window=[-60, 50])
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram, self.binned_st1,
            self.binned_st2, window=[-50, 60])
        # Test for no integer or wrong string in input
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram, self.binned_st1,
            self.binned_st2, window=[-25.5, 25.5])
        self.assertRaises(
            ValueError, sc.cross_correlation_histogram, self.binned_st1,
            self.binned_st2, window='test')

    def test_border_correction(self):
        """Test if the border correction for bins at the edges is correctly
        performed"""

        # check that nothing changes for valid lags
        cch_valid, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='full',
            border_correction=True, binary=False, kernel=None)
        valid_lags = sc._CrossCorrHist.get_valid_lags(self.binned_st1,
                                                      self.binned_st2)
        left_edge, right_edge = valid_lags[(0, -1), ]
        cch_builder = sc._CrossCorrHist(self.binned_st1, self.binned_st2,
                                        window=(left_edge, right_edge))
        cch_valid = cch_builder.correlate_speed(cch_mode='valid')
        cch_corrected = cch_builder.border_correction(cch_valid)

        np.testing.assert_array_equal(cch_valid, cch_corrected)

        # test the border correction for lags without full overlap
        cch_full, lags_full = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='full')

        cch_full_corrected, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, window='full',
            border_correction=True)

        n_bins_outside_window = np.min(np.abs(
            np.subtract.outer(lags_full, valid_lags)), axis=1)

        min_n_bins = min(self.binned_st1.n_bins, self.binned_st2.n_bins)

        border_correction = (cch_full_corrected / cch_full).magnitude.flatten()

        # exclude NaNs caused by zeros in the cch
        mask = np.logical_not(np.isnan(border_correction))

        np.testing.assert_array_almost_equal(
            border_correction[mask],
            (float(min_n_bins)
             / (min_n_bins - n_bins_outside_window))[mask])

    def test_kernel(self):
        """Test if the smoothing kernel is correctly defined, and wheter it is
        applied properly."""
        smoothed_cch, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, kernel=np.ones(3))
        smoothed_cch_mem, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, kernel=np.ones(3),
            method='memory')

        cch, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, kernel=None)
        cch_mem, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, kernel=None, method='memory')

        self.assertNotEqual(smoothed_cch.all, cch.all)
        self.assertNotEqual(smoothed_cch_mem.all, cch_mem.all)

        self.assertRaises(
            ValueError, sc.cch, self.binned_st1, self.binned_st2,
            kernel=np.ones(100))
        self.assertRaises(
            ValueError, sc.cch, self.binned_st1, self.binned_st2,
            kernel=np.ones(100), method='memory')

    def test_exist_alias(self):
        """
        Test if alias cch still exists.
        """
        self.assertEqual(sc.cross_correlation_histogram, sc.cch)

    def test_annotations(self):
        cch, _ = sc.cross_correlation_histogram(
            self.binned_st1, self.binned_st2, kernel=np.ones(3))
        target_dict = dict(window='full', border_correction=False,
                           binary=False, kernel=True,
                           normalization='counts')
        self.assertIn('cch_parameters', cch.annotations)
        self.assertEqual(cch.annotations['cch_parameters'], target_dict)


class CrossCorrelationHistDifferentTStartTStopTest(unittest.TestCase):

    def _run_sub_tests(self, st1, st2, lags_true):
        for window in ('valid', 'full'):
            for method in ('speed', 'memory'):
                with self.subTest(window=window, method=method):
                    bin_size = 1 * pq.s
                    st1_binned = conv.BinnedSpikeTrain(st1, bin_size=bin_size)
                    st2_binned = conv.BinnedSpikeTrain(st2, bin_size=bin_size)
                    left, right = lags_true[window][(0, -1), ]
                    cch_window, lags_window = sc.cross_correlation_histogram(
                        st1_binned, st2_binned, window=(left, right),
                        method=method,
                    )
                    cch, lags = sc.cross_correlation_histogram(
                        st1_binned, st2_binned, window=window)

                    # target cross correlation
                    cch_target = np.correlate(st1_binned.to_array()[0],
                                              st2_binned.to_array()[0],
                                              mode=window)

                    self.assertEqual(len(lags_window), cch_window.shape[0])
                    assert_array_almost_equal(cch.magnitude,
                                              cch_window.magnitude)
                    # the output is reversed since we cross-correlate
                    # st2 with st1 rather than st1 with st2 (numpy behavior)
                    assert_array_almost_equal(np.ravel(cch.magnitude),
                                              cch_target[::-1])
                    assert_array_equal(lags, lags_true[window])
                    assert_array_equal(lags, lags_window)

    def test_cross_correlation_histogram_valid_full_overlap(self):
        # ex. 1 in the source code
        st1 = neo.SpikeTrain([3.5, 4.5, 7.5] * pq.s, t_start=3 * pq.s,
                             t_stop=8 * pq.s)
        st2 = neo.SpikeTrain([1.5, 2.5, 4.5, 8.5, 9.5, 10.5]
                             * pq.s, t_start=1 * pq.s, t_stop=13 * pq.s)
        lags_true = {
            'valid': np.arange(-2, 6, dtype=np.int32),
            'full': np.arange(-6, 10, dtype=np.int32)
        }
        self._run_sub_tests(st1, st2, lags_true)

    def test_cross_correlation_histogram_valid_partial_overlap(self):
        # ex. 2 in the source code
        st1 = neo.SpikeTrain([2.5, 3.5, 4.5, 6.5] * pq.s, t_start=1 * pq.s,
                             t_stop=7 * pq.s)
        st2 = neo.SpikeTrain([3.5, 5.5, 6.5, 7.5, 8.5] *
                             pq.s, t_start=2 * pq.s, t_stop=9 * pq.s)
        lags_true = {
            'valid': np.arange(1, 3, dtype=np.int32),
            'full': np.arange(-4, 8, dtype=np.int32)
        }
        self._run_sub_tests(st1, st2, lags_true)

    def test_cross_correlation_histogram_valid_no_overlap(self):
        st1 = neo.SpikeTrain([2.5, 3.5, 4.5, 6.5] * pq.s, t_start=1 * pq.s,
                             t_stop=7 * pq.s)
        st2 = neo.SpikeTrain([3.5, 5.5, 6.5, 7.5, 8.5] * pq.s + 6 * pq.s,
                             t_start=8 * pq.s, t_stop=15 * pq.s)
        lags_true = {
            'valid': np.arange(7, 9, dtype=np.int32),
            'full': np.arange(2, 14, dtype=np.int32)
        }
        self._run_sub_tests(st1, st2, lags_true)

    def test_invalid_time_shift(self):
        # time shift of 0.4 s is not multiple of bin_size=1 s
        st1 = neo.SpikeTrain([2.5, 3.5] * pq.s, t_start=1 * pq.s,
                             t_stop=7 * pq.s)
        st2 = neo.SpikeTrain([3.5, 5.5] * pq.s, t_start=1.4 * pq.s,
                             t_stop=7.4 * pq.s)
        bin_size = 1 * pq.s
        st1_binned = conv.BinnedSpikeTrain(st1, bin_size=bin_size)
        st2_binned = conv.BinnedSpikeTrain(st2, bin_size=bin_size)
        self.assertRaises(ValueError, sc.cross_correlation_histogram,
                          st1_binned, st2_binned)


class SpikeTimeTilingCoefficientTestCase(unittest.TestCase):

    def setUp(self):
        # These two arrays must be such that they do not have coincidences
        # spanning across two neighbor bins assuming ms bins [0,1),[1,2),...
        self.test_array_1d_1 = [
            1.3, 7.56, 15.87, 28.23, 30.9, 34.2, 38.2, 43.2]
        self.test_array_1d_2 = [
            1.02, 2.71, 18.82, 28.46, 28.79, 43.6]

        # Build spike trains
        self.st_1 = neo.SpikeTrain(
            self.test_array_1d_1, units='ms', t_stop=50.)
        self.st_2 = neo.SpikeTrain(
            self.test_array_1d_2, units='ms', t_stop=50.)

    def test_sttc_dt_smaller_zero(self):
        self.assertRaises(ValueError, sc.sttc, self.st_1, self.st_2,
                          dt=0 * pq.s)
        self.assertRaises(ValueError, sc.sttc, self.st_1, self.st_2,
                          dt=-1 * pq.ms)

    def test_sttc_different_t_stop(self):
        st_1 = neo.SpikeTrain([1], units='ms', t_stop=10.)
        st_2 = neo.SpikeTrain([5], units='ms', t_stop=10.)
        st_2.t_stop = 1 * pq.ms
        self.assertRaises(ValueError, sc.sttc, st_1, st_2)

    def test_sttc_different_t_start(self):
        st_1 = neo.SpikeTrain([1], units='ms', t_stop=10.)
        st_2 = neo.SpikeTrain([5], units='ms', t_stop=10.)
        st_2.t_start = 1 * pq.ms
        self.assertRaises(ValueError, sc.sttc, st_1, st_2)

    def test_sttc_different_units_dt(self):
        # test for result
        # target obtained with pencil and paper according to original paper.
        target = 0.495860165593
        self.assertAlmostEqual(target, sc.sttc(self.st_1, self.st_2,
                                               0.005 * pq.s))

        # test for same result with dt given in ms
        self.assertAlmostEqual(target, sc.sttc(self.st_1, self.st_2,
                                               5.0 * pq.ms))

    def test_sttc_different_units_spiketrains(self):
        st1 = neo.SpikeTrain([1], units='ms', t_stop=10.)
        st2 = neo.SpikeTrain([5], units='s', t_stop=10.)
        self.assertRaises(ValueError, sc.sttc, st1, st2)

    def test_sttc_not_enough_spiketrains(self):
        # test no spiketrains
        self.assertRaises(TypeError, sc.sttc, [], [])

        # test one spiketrain
        self.assertRaises(TypeError, sc.sttc, self.st_1, [])

    def test_sttc_one_spike(self):
        # test for one spike in a spiketrain
        st1 = neo.SpikeTrain([1], units='ms', t_stop=10.)
        st2 = neo.SpikeTrain([5], units='ms', t_stop=10.)
        self.assertEqual(sc.sttc(st1, st2), 1.0)
        self.assertTrue(bool(sc.sttc(st1, st2, 0.1 * pq.ms) < 0))

    def test_sttc_high_value_dt(self):
        # test for high value of dt
        self.assertEqual(sc.sttc(self.st_1, self.st_2, dt=5 * pq.s), 1.0)

    def test_sttc_edge_cases(self):
        # test for TA = PB = 1 but TB /= PA /= 1 and vice versa
        st2 = neo.SpikeTrain([5], units='ms', t_stop=10.)
        st3 = neo.SpikeTrain([1, 5, 9], units='ms', t_stop=10.)
        target2 = 1. / 3.

        self.assertAlmostEqual(target2, sc.sttc(st3, st2, 0.003 * pq.s))
        self.assertAlmostEqual(target2, sc.sttc(st2, st3, 0.003 * pq.s))

    def test_sttc_unsorted_spiketimes(self):
        # regression test for issue  #563
        # https://github.com/NeuralEnsemble/elephant/issues/563
        spiketrain_E7 = neo.SpikeTrain(
            [1678., 23786.3, 34641.8, 71520.7, 73606.9, 78383.3,
             97387.9, 144313.4, 4607.6, 19275.1, 152894.2, 44240.1],
            units='ms', t_stop=300000 * pq.ms)

        spiketrain_E3 = neo.SpikeTrain(
            [1678., 23786.3, 34641.8, 71520.7, 73606.9, 78383.3,
             97387.9, 144313.4, 4607.6, 19275.1, 152894.2, 44240.1],
            units='ms', t_stop=300000 * pq.ms)
        sttc_unsorted_E7_E3 = sc.sttc(spiketrain_E7,
                                      spiketrain_E3, dt=0.10 * pq.s)
        self.assertAlmostEqual(sttc_unsorted_E7_E3, 1)
        spiketrain_E7.sort()
        spiketrain_E3.sort()
        sttc_sorted_E7_E3 = sc.sttc(spiketrain_E7,
                                    spiketrain_E3, dt=0.10 * pq.s)
        self.assertAlmostEqual(sttc_unsorted_E7_E3, sttc_sorted_E7_E3)

        spiketrain_E8 = neo.SpikeTrain(
            [20646.8, 25875.1, 26154.4, 35121., 55909.7, 79164.8,
             110849.8, 117484.1, 3731.5, 4213.9, 119995.1, 123748.1,
             171016.8, 172989., 185145.2, 12043.5, 185995.9, 186740.1,
             12629.8, 23394.3, 34993.2], units='ms', t_stop=300000 * pq.ms)

        spiketrain_B3 = neo.SpikeTrain(
            [10600.7, 19699.6, 22803., 40769.3, 121385.7, 127402.9,
             130829.2, 134363.8, 1193.5, 8012.7, 142037.3, 146628.2,
             165925.3, 168489.3, 175194.3, 10339.8, 178676.4, 180807.2,
             201431.3, 22231.1, 38113.4], units='ms', t_stop=300000 * pq.ms)

        self.assertTrue(
            sc.sttc(spiketrain_E8, spiketrain_B3, dt=0.10 * pq.s) < 1)

        sttc_unsorted_E8_B3 = sc.sttc(spiketrain_E8,
                                      spiketrain_B3, dt=0.10 * pq.s)
        spiketrain_E8.sort()
        spiketrain_B3.sort()
        sttc_sorted_E8_B3 = sc.sttc(spiketrain_E8,
                                    spiketrain_B3, dt=0.10 * pq.s)
        self.assertAlmostEqual(sttc_unsorted_E8_B3, sttc_sorted_E8_B3)

    def test_sttc_validation_test(self):
        """This test checks the results of elephants implementation of
        the spike time tiling coefficient against the results of the
        original c-implementation.
        The c-code and the test data is located at
        NeuralEnsemble/elephant-data/unittest/spike_train_correlation/
        spike_time_tiling_coefficient"""

        repo_path = r"unittest/spike_train_correlation/spike_time_tiling_coefficient/data"  # noqa

        files_to_download = [("spike_time_tiling_coefficient_results.nix",
                              "e3749d79046622494660a03e89950f51")]

        downloaded_files = {}
        for filename, checksum in files_to_download:
            downloaded_files[filename] = {
                'filename': filename,
                'path': download_datasets(repo_path=f"{repo_path}/{filename}",
                                          checksum=checksum)}

        reader = NixIO(downloaded_files[
            'spike_time_tiling_coefficient_results.nix']['path'], mode='ro')
        test_data_block = reader.read()

        for segment in test_data_block[0].segments:
            spiketrain_i = segment.spiketrains[0]
            spiketrain_j = segment.spiketrains[1]
            dt = segment.annotations['dt']
            sttc_result = segment.annotations['sttc_result']
            self.assertAlmostEqual(sc.sttc(spiketrain_i, spiketrain_j, dt),
                                   sttc_result)

    def test_sttc_exist_alias(self):
        # Test if alias cch still exists.
        self.assertEqual(sc.spike_time_tiling_coefficient, sc.sttc)


class SpikeTrainTimescaleTestCase(unittest.TestCase):

    def test_timescale_calculation(self):
        """
        Test the timescale generation using an alpha-shaped ISI distribution,
        see [1, eq. 1.68]. This is equivalent to a homogeneous gamma process
        with alpha=2 and beta=2*nu where nu is the rate.

        For this process, the autocorrelation function is given by a sum of a
        delta peak and a (negative) exponential, see [1, eq. 1.69].
        The exponential decays with \tau_corr = 1 / (4*nu), thus this fixes
        timescale.

        [1] Lindner, B. (2009). A brief introduction to some simple stochastic
            processes. Stochastic Methods in Neuroscience, 1.
        """
        nu = 25 / pq.s
        T = 15 * pq.min
        bin_size = 1 * pq.ms
        timescale = 1 / (4 * nu)
        np.random.seed(35)

        for _ in range(10):
            spikes = StationaryGammaProcess(rate=2 * nu / 2, shape_factor=2,
                                            t_start=0 * pq.ms,
                                            t_stop=T).generate_spiketrain()
            spikes_bin = conv.BinnedSpikeTrain(spikes, bin_size)
            timescale_i = sc.spike_train_timescale(spikes_bin, 10 * timescale)
            assert_array_almost_equal(timescale, timescale_i, decimal=3)

    def test_timescale_errors(self):
        spikes = neo.SpikeTrain([1, 5, 7, 8] * pq.ms, t_stop=10 * pq.ms)
        binsize = 1 * pq.ms
        spikes_bin = conv.BinnedSpikeTrain(spikes, binsize)

        # Tau max with no units
        tau_max = 1
        self.assertRaises(ValueError,
                          sc.spike_train_timescale, spikes_bin, tau_max)

        # Tau max that is not a multiple of the binsize
        tau_max = 1.1 * pq.ms
        self.assertRaises(ValueError,
                          sc.spike_train_timescale, spikes_bin, tau_max)

    def test_timescale_nan(self):
        st0 = neo.SpikeTrain([] * pq.ms, t_stop=10 * pq.ms)
        st1 = neo.SpikeTrain([1] * pq.ms, t_stop=10 * pq.ms)
        st2 = neo.SpikeTrain([1, 5] * pq.ms, t_stop=10 * pq.ms)
        st3 = neo.SpikeTrain([1, 5, 6] * pq.ms, t_stop=10 * pq.ms)
        st4 = neo.SpikeTrain([1, 5, 6, 9] * pq.ms, t_stop=10 * pq.ms)

        binsize = 1 * pq.ms
        tau_max = 1 * pq.ms

        for st in [st0, st1]:
            bst = conv.BinnedSpikeTrain(st, binsize)
            with self.assertWarns(UserWarning):
                timescale = sc.spike_train_timescale(bst, tau_max)
            self.assertTrue(math.isnan(timescale))

        for st in [st2, st3, st4]:
            bst = conv.BinnedSpikeTrain(st, binsize)
            timescale = sc.spike_train_timescale(bst, tau_max)
            self.assertFalse(math.isnan(timescale))


if __name__ == '__main__':
    unittest.main()



"""
The ``elephant.parallel`` package provides classes to parallelize calls to any
user-specified function.

The typical use case is calling a function many times with different
parameters.

.. note::  This parallelization module is independent from Elephant and can be
           used to parallelize mutually independent calls to arbitrary
           functions.

Tutorial
--------

:doc:`View tutorial <../tutorials/parallel>`

Run tutorial interactively:

.. image:: https://mybinder.org/badge.svg
   :target: https://mybinder.org/v2/gh/NeuralEnsemble/elephant/master
            ?filepath=doc/tutorials/parallel.ipynb


Available Executors
-------------------

.. autosummary::
    :toctree: _toctree/parallel/

    ProcessPoolExecutor
    MPIPoolExecutor
    MPICommExecutor


:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import warnings

from .parallel import ProcessPoolExecutor, SingleProcess

try:
    from .mpi import MPIPoolExecutor, MPICommExecutor
except ImportError:
    # mpi4py is missing
    warnings.warn("mpi4py package is missing. Please run 'pip install mpi4py' "
                  "in a terminal to activate MPI features.")

__all__ = [
    "ProcessPoolExecutor",
    "SingleProcess",
    "MPIPoolExecutor",
    "MPICommExecutor"
]



# -*- coding: utf-8 -*-
"""
Unit tests for the spectral module.

:copyright: Copyright 2014-2024 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import unittest

import neo.core
import numpy as np
import pytest
import quantities as pq
import scipy
import scipy.fft
import scipy.signal as spsig
from neo import AnalogSignal
from numpy.testing import assert_array_equal
from packaging import version

import elephant.spectral
from elephant.datasets import download_datasets


class WelchPSDTestCase(unittest.TestCase):
    def test_welch_psd_errors(self):
        # generate a dummy data
        data = AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                            units='mV')

        # check for invalid parameter values
        # - length of segments
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          len_segment=0)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          len_segment=data.shape[0] * 2)
        # - number of segments
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          n_segments=0)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          n_segments=data.shape[0] * 2)
        # - frequency resolution
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          frequency_resolution=-1)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          frequency_resolution=data.sampling_rate /
                          (data.shape[0] + 1))
        # - overlap
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          overlap=-1.0)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          overlap=1.1)

    def test_welch_psd_warnings(self):
        # generate a dummy data
        data = AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                            units='mV')
        # Test deprecation warning for 'hanning' window
        self.assertWarns(DeprecationWarning, elephant.spectral.welch_psd,
                         data, window='hanning')

    def test_welch_psd_behavior(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=data_length)
        signal = [np.sin(2 * np.pi * signal_freq * t)
                  for t in np.arange(0, data_length * sampling_period,
                                     sampling_period)]
        data = AnalogSignal(np.array(signal + noise),
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # consistency between different ways of specifying segment length
        freqs1, psd1 = elephant.spectral.welch_psd(
            data, len_segment=data_length // 5, overlap=0)
        freqs2, psd2 = elephant.spectral.welch_psd(
            data, n_segments=5, overlap=0)
        self.assertTrue((psd1 == psd2).all() and (freqs1 == freqs2).all())

        # frequency resolution and consistency with data
        freq_res = 1.0 * pq.Hz
        freqs, psd = elephant.spectral.welch_psd(
            data, frequency_resolution=freq_res)
        self.assertAlmostEqual(freq_res, freqs[1] - freqs[0])
        self.assertEqual(freqs[psd.argmax()], signal_freq)
        freqs_np, psd_np = elephant.spectral.welch_psd(
            data.magnitude.flatten(), fs=1 / sampling_period,
            frequency_resolution=freq_res)
        self.assertTrue((freqs == freqs_np).all() and (psd == psd_np).all())

        # check of scipy.signal.welch() parameters
        params = {'window': 'hamming', 'nfft': 1024, 'detrend': 'linear',
                  'return_onesided': False, 'scaling': 'spectrum'}
        for key, val in params.items():
            freqs, psd = elephant.spectral.welch_psd(
                data, len_segment=1000, overlap=0, **{key: val})
            freqs_spsig, psd_spsig = spsig.welch(np.rollaxis(data, 0, len(
                data.shape)), fs=1 / sampling_period, nperseg=1000,
                                                 noverlap=0, **{key: val})
            self.assertTrue(
                (freqs == freqs_spsig).all() and (
                        psd == psd_spsig).all())

        # - generate multidimensional data for check of parameter `axis`
        num_channel = 4
        data_length = 5000
        data_multidim = np.random.normal(size=(num_channel, data_length))
        freqs, psd = elephant.spectral.welch_psd(data_multidim)
        freqs_T, psd_T = elephant.spectral.welch_psd(data_multidim.T, axis=0)
        self.assertTrue(np.all(freqs == freqs_T))
        self.assertTrue(np.all(psd == psd_T.T))

    def test_welch_psd_input_types(self):
        # generate a test data
        sampling_period = 0.001
        data = AnalogSignal(np.array(np.random.normal(size=5000)),
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # outputs from AnalogSignal input are of Quantity type (standard usage)
        freqs_neo, psd_neo = elephant.spectral.welch_psd(data)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, psd_pq = elephant.spectral.welch_psd(
            data.magnitude.flatten() * data.units, fs=1 / sampling_period)
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, psd_np = elephant.spectral.welch_psd(
            data.magnitude.flatten(), fs=1 / sampling_period)
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(psd_np, pq.quantity.Quantity))

        # check if the results from different input types are identical
        self.assertTrue(
            (freqs_neo == freqs_pq).all() and (
                    psd_neo == psd_pq).all())
        self.assertTrue(
            (freqs_neo == freqs_np).all() and (
                    psd_neo == psd_np).all())

    def test_welch_psd_multidim_input(self):
        # generate multidimensional data
        num_channel = 4
        data_length = 5000
        sampling_period = 0.001
        noise = np.random.normal(size=(num_channel, data_length))
        data_np = np.array(noise)
        # Since row-column order in AnalogSignal is different from the
        # conventional one, `data_np` needs to be transposed when it's used to
        # define an AnalogSignal
        data_neo = AnalogSignal(data_np.T,
                                sampling_period=sampling_period * pq.s,
                                units='mV')
        data_neo_1dim = AnalogSignal(data_np[0],
                                     sampling_period=sampling_period * pq.s,
                                     units='mV')

        # check if the results from different input types are identical
        freqs_np, psd_np = elephant.spectral.welch_psd(data_np,
                                                       fs=1 / sampling_period)
        freqs_neo, psd_neo = elephant.spectral.welch_psd(data_neo)
        freqs_neo_1dim, psd_neo_1dim = elephant.spectral.welch_psd(
            data_neo_1dim)
        self.assertTrue(np.all(freqs_np == freqs_neo))
        self.assertTrue(np.all(psd_np == psd_neo))
        self.assertTrue(np.all(psd_neo_1dim == psd_neo[0]))


class MultitaperPSDTestCase(unittest.TestCase):
    def test_multitaper_psd_errors(self):
        # generate dummy data
        data_length = 5000
        signal = AnalogSignal(np.zeros(data_length),
                              sampling_period=0.001 * pq.s,
                              units='mV')
        fs = signal.sampling_rate
        self.assertIsInstance(fs, pq.Quantity)

        # check for invalid parameter values
        # - number of tapers
        self.assertRaises(ValueError, elephant.spectral.multitaper_psd, signal,
                          num_tapers=-5)
        self.assertRaises(TypeError, elephant.spectral.multitaper_psd, signal,
                          num_tapers=-5.0)
        # - peak resolution
        self.assertRaises(ValueError, elephant.spectral.multitaper_psd, signal,
                          peak_resolution=-1)

    def test_multitaper_psd_behavior(self):
        # generate data (frequency domain to time domain)
        r = np.ones(2501) * 0.2
        r[0], r[500] = 0, 10  # Zero DC, peak at 100 Hz
        phi = np.random.uniform(-np.pi, np.pi, len(r))
        fake_coeffs = r*np.exp(1j * phi)
        fake_ts = scipy.fft.irfft(fake_coeffs)
        sampling_period = 0.001
        freqs = scipy.fft.rfftfreq(len(fake_ts), d=sampling_period)
        signal_freq = freqs[r.argmax()]

        data = AnalogSignal(fake_ts, sampling_period=sampling_period * pq.s,
                            units='mV')

        # consistency between different ways of specifying number of tapers
        freqs1, psd1 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3.5)
        freqs2, psd2 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3.5,
                                                        num_tapers=6)
        self.assertTrue((psd1 == psd2).all() and (freqs1 == freqs2).all())

        # peak resolution and consistency with data
        peak_res = 1.0 * pq.Hz
        freqs, psd = elephant.spectral.multitaper_psd(
            data, peak_resolution=peak_res)
        self.assertEqual(freqs[psd.argmax()], signal_freq)
        freqs_np, psd_np = elephant.spectral.multitaper_psd(
            data.magnitude.flatten(), fs=1 / sampling_period,
            peak_resolution=peak_res)
        self.assertTrue((freqs == freqs_np).all() and (psd == psd_np).all())

    def test_multitaper_psd_parameter_hierarchy(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=data_length)
        signal = [np.sin(2 * np.pi * signal_freq * t)
                  for t in np.arange(0, data_length * sampling_period,
                                     sampling_period)]
        data = AnalogSignal(np.array(signal + noise),
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # Test num_tapers vs nw
        freqs1, psd1 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3,
                                                        num_tapers=9)
        freqs2, psd2 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3)
        self.assertTrue((freqs1 == freqs2).all() and (psd1 != psd2).all())

        # Test peak_resolution vs nw
        freqs1, psd1 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3,
                                                        num_tapers=9,
                                                        peak_resolution=1)
        freqs2, psd2 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3,
                                                        num_tapers=9)
        self.assertTrue((freqs1 == freqs2).all() and (psd1 != psd2).all())

    def test_multitaper_psd_against_nitime(self):
        """
        This test assesses the match between this implementation of
        multitaper against nitime (0.8) using a predefined time series
        generated by an autoregressive model.

        Please follow the link below for more details:
        https://gin.g-node.org/INM-6/elephant-data/src/master/unittest/spectral/multitaper_psd
        """
        repo_path = r"unittest/spectral/multitaper_psd/data"

        files_to_download = [
            ("time_series.npy", "ff43797e2ac94613f510b20a31e2e80e"),
            ("psd_nitime.npy", "89d1f53957e66c786049ea425b53c0e8")
        ]

        downloaded_files = {}
        for filename, checksum in files_to_download:
            downloaded_files[filename] = {
                'filename': filename,
                'path': download_datasets(repo_path=f"{repo_path}/{filename}",
                                          checksum=checksum)}

        time_series = np.load(downloaded_files['time_series.npy']['path'])
        psd_nitime = np.load(downloaded_files['psd_nitime.npy']['path'])

        freqs, psd_multitaper = elephant.spectral.multitaper_psd(
            signal=time_series, fs=0.1, nw=4, num_tapers=8)

        np.testing.assert_allclose(np.squeeze(psd_multitaper), psd_nitime,
                                   rtol=0.3, atol=0.1)

    def test_multitaper_psd_input_types(self):
        # generate a test data
        sampling_period = 0.001
        data = AnalogSignal(np.array(np.random.normal(size=5000)),
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # outputs from AnalogSignal input are of Quantity type (standard usage)
        freqs_neo, psd_neo = elephant.spectral.multitaper_psd(data)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, psd_pq = elephant.spectral.multitaper_psd(
            data.magnitude.flatten() * data.units, fs=1 / sampling_period)
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, psd_np = elephant.spectral.multitaper_psd(
            data.magnitude.flatten(), fs=1 / sampling_period)
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(psd_np, pq.quantity.Quantity))

        # fs with and without units
        fs_hz = 1 * pq.Hz
        fs_int = 1
        freqs_fs_hz, psd_fs_hz = elephant.spectral.multitaper_psd(
            data.magnitude.T, fs=fs_hz)
        freqs_fs_int, psd_fs_int = elephant.spectral.multitaper_psd(
            data.magnitude.T, fs=fs_int)

        np.testing.assert_array_equal(freqs_fs_hz, freqs_fs_int)
        np.testing.assert_array_equal(psd_fs_hz, psd_fs_int)

        # check if the results from different input types are identical
        self.assertTrue(
            (freqs_neo == freqs_pq).all() and (
                    psd_neo == psd_pq).all())
        self.assertTrue(
            (freqs_neo == freqs_np).all() and (
                    psd_neo == psd_np).all())


class SegmentedMultitaperPSDTestCase(unittest.TestCase):
    # The following assertions test _segmented_apply_func in the context
    # of segmented_multitaper_psd. In other words, only the segmentation is
    # addressed. The inner workings of the multitaper function are tested
    # in the MultitaperPSDTestCase.
    def test_segmented_multitaper_psd_errors(self):
        # generate dummy data
        data_length = 5000
        signal = AnalogSignal(np.zeros(data_length),
                              sampling_period=0.001 * pq.s,
                              units='mV')
        fs = signal.sampling_rate

        # check for invalid parameter values
        # - frequency resolution
        self.assertRaises(ValueError,
                          elephant.spectral.segmented_multitaper_psd, signal,
                          frequency_resolution=-10)

        # - n per segment
        # n_per_seg = int(fs / dF), where dF is the frequency_resolution
        broken_freq_resolution = fs / (data_length+1)
        self.assertRaises(ValueError,
                          elephant.spectral.segmented_multitaper_psd, signal,
                          frequency_resolution=broken_freq_resolution)

        # - length of segment (negative)
        self.assertRaises(ValueError,
                          elephant.spectral.segmented_multitaper_psd, signal,
                          len_segment=-10)

        # - length of segment (larger than data length)
        self.assertRaises(ValueError,
                          elephant.spectral.segmented_multitaper_psd, signal,
                          len_segment=data_length+1)

        # - number of segments (negative)
        self.assertRaises(ValueError,
                          elephant.spectral.segmented_multitaper_psd, signal,
                          n_segments=-10)

        # - number of segments (larger than data length)
        self.assertRaises(ValueError,
                          elephant.spectral.segmented_multitaper_psd, signal,
                          n_segments=data_length+1)

    def test_segmented_multitaper_psd_behavior(self):
        # generate data (frequency domain to time domain)
        r = np.ones(2501) * 0.2
        r[0], r[500] = 0, 10  # Zero DC, peak at 100 Hz
        phi = np.random.uniform(-np.pi, np.pi, len(r))
        fake_coeffs = r*np.exp(1j * phi)
        fake_ts = scipy.fft.irfft(fake_coeffs)
        sampling_period = 0.001
        freqs = scipy.fft.rfftfreq(len(fake_ts), d=sampling_period)
        signal_freq = freqs[r.argmax()]

        data = AnalogSignal(fake_ts, sampling_period=sampling_period * pq.s,
                            units='mV')

        # consistency between different ways of specifying n_per_seg
        # n_per_seg = int(fs/dF) and n_per_seg = len_segment
        frequency_resolution = 1 * pq.Hz
        len_segment = int(data.sampling_rate / frequency_resolution)

        freqs_fr, psd_fr = elephant.spectral.segmented_multitaper_psd(
            data, frequency_resolution=frequency_resolution)

        freqs_ls, psd_ls = elephant.spectral.segmented_multitaper_psd(
            data, len_segment=len_segment)

        np.testing.assert_array_equal(freqs_fr, freqs_ls)
        np.testing.assert_array_equal(psd_fr, psd_ls)

    def test_segmented_multitaper_psd_parameter_hierarchy(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=data_length)
        signal = [np.sin(2 * np.pi * signal_freq * t)
                  for t in np.arange(0, data_length * sampling_period,
                                     sampling_period)]
        data = AnalogSignal(np.array(signal + noise),
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # test frequency_resolution vs len_segment vs n_segments
        n_segments = 5
        len_segment = 2000
        frequency_resolution = 0.25 * pq.Hz

        freqs_ns, psd_ns = \
            elephant.spectral.segmented_multitaper_psd(
                data, n_segments=n_segments)

        freqs_ls, psd_ls = \
            elephant.spectral.segmented_multitaper_psd(
                data, n_segments=n_segments, len_segment=len_segment)

        freqs_fr, psd_fr = \
            elephant.spectral.segmented_multitaper_psd(
                data, n_segments=n_segments, len_segment=len_segment,
                frequency_resolution=frequency_resolution)

        self.assertTrue(freqs_ns.shape < freqs_ls.shape < freqs_fr.shape)
        self.assertTrue(psd_ns.shape < psd_ls.shape < psd_fr.shape)

    def test_segmented_multitaper_psd_input_types(self):
        # generate a test data
        sampling_period = 0.001
        data = AnalogSignal(np.array(np.random.normal(size=5000)),
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # outputs from AnalogSignal input are of Quantity type (standard usage)
        freqs_neo, psd_neo = elephant.spectral.segmented_multitaper_psd(data)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, psd_pq = elephant.spectral.segmented_multitaper_psd(
            data.magnitude.flatten() * data.units, fs=1 / sampling_period)
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, psd_np = elephant.spectral.segmented_multitaper_psd(
            data.magnitude.flatten(), fs=1 / sampling_period)
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(psd_np, pq.quantity.Quantity))

        # frequency resolution with and without units
        freq_res_hz = 1 * pq.Hz
        freq_res_int = 1

        freqs_int, psd_int = elephant.spectral.segmented_multitaper_psd(
            data, frequency_resolution=freq_res_int)

        freqs_hz, psd_hz = elephant.spectral.segmented_multitaper_psd(
            data, frequency_resolution=freq_res_hz)

        np.testing.assert_array_equal(freqs_int, freqs_hz)
        np.testing.assert_array_equal(psd_int, psd_hz)

        # fs with and without units
        fs_hz = 1 * pq.Hz
        fs_int = 1
        freqs_fs_hz, psd_fs_hz = elephant.spectral.multitaper_psd(
            data.magnitude.T, fs=fs_hz)
        freqs_fs_int, psd_fs_int = elephant.spectral.multitaper_psd(
            data.magnitude.T, fs=fs_int)

        np.testing.assert_array_equal(freqs_fs_hz, freqs_fs_int)
        np.testing.assert_array_equal(psd_fs_hz, psd_fs_int)

        # check if the results from different input types are identical
        self.assertTrue(
            (freqs_neo == freqs_pq).all() and (
                    psd_neo == psd_pq).all())
        self.assertTrue(
            (freqs_neo == freqs_np).all() and (
                    psd_neo == psd_np).all())


class MultitaperCrossSpectrumTestCase(unittest.TestCase):
    def test_multitaper_cross_spectrum_errors(self):
        # generate dummy data
        data_length = 5000
        signal = AnalogSignal(np.zeros(data_length),
                              sampling_period=0.001 * pq.s,
                              units='mV')
        fs = signal.sampling_rate

        # check for invalid parameter values
        # - number of tapers
        self.assertRaises(ValueError,
                          elephant.spectral.multitaper_cross_spectrum, signal,
                          fs=fs, num_tapers=-5)
        self.assertRaises(TypeError,
                          elephant.spectral.multitaper_cross_spectrum, signal,
                          fs=fs, num_tapers=-5.0)

        # - peak resolution
        self.assertRaises(ValueError,
                          elephant.spectral.multitaper_cross_spectrum, signal,
                          fs=fs, peak_resolution=-1)

    def test_multitaper_cross_spectrum_behavior(self):
        # generate data (frequency domain to time domain)
        r = np.ones(2501) * 0.2
        r[0], r[500] = 0, 10  # Zero DC, peak at 100 Hz
        phi_x = np.random.uniform(-np.pi, np.pi, len(r))
        phi_y = np.random.uniform(-np.pi, np.pi, len(r))
        fake_coeffs_x = r*np.exp(1j * phi_x)
        fake_coeffs_y = r*np.exp(1j * phi_y)
        signal_x = scipy.fft.irfft(fake_coeffs_x)
        signal_y = scipy.fft.irfft(fake_coeffs_y)
        sampling_period = 0.001
        freqs = scipy.fft.rfftfreq(len(signal_x), d=sampling_period)
        signal_freq = freqs[r.argmax()]
        data = AnalogSignal(np.vstack([signal_x, signal_y]).T,
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # consistency between different ways of specifying number of tapers
        freqs1, cross_spec1 = \
            elephant.spectral.multitaper_cross_spectrum(
                    data,
                    fs=data.sampling_rate,
                    nw=3.5)
        freqs2, cross_spec2 = \
            elephant.spectral.multitaper_cross_spectrum(
                    data,
                    fs=data.sampling_rate,
                    nw=3.5,
                    num_tapers=6)
        self.assertTrue((cross_spec1 == cross_spec2).all()
                        and (freqs1 == freqs2).all())

        # peak resolution and consistency with data
        peak_res = 1.0 * pq.Hz
        freqs, cross_spec = \
            elephant.spectral.multitaper_cross_spectrum(
                    data, peak_resolution=peak_res)

        self.assertEqual(freqs[cross_spec[0, 0].argmax()], signal_freq)
        freqs_np, cross_spec_np = \
            elephant.spectral.multitaper_cross_spectrum(
                    data.magnitude.T, fs=1 / sampling_period,
                    peak_resolution=peak_res)
        self.assertTrue((freqs == freqs_np).all()
                        and (cross_spec == cross_spec_np).all())

        # one-sided vs two-sided spectrum
        freqs_os, cross_spec_os = \
            elephant.spectral.multitaper_cross_spectrum(
                data, return_onesided=True)

        freqs_ts, cross_spec_ts = \
            elephant.spectral.multitaper_cross_spectrum(
                data, return_onesided=False)

        # Nyquist frequency is negative when using onesided=False (fftfreq)
        # See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.fft.rfftfreq.html#scipy.fft.rfftfreq  # noqa
        nonnegative_freqs_indices = np.nonzero(freqs_ts >= 0)[0]
        nyquist_freq_idx = np.abs(freqs_ts).argmax()
        ts_freq_indices = np.append(nonnegative_freqs_indices,
                                    nyquist_freq_idx)
        ts_overlap_freqs = np.append(
            freqs_ts[nonnegative_freqs_indices].rescale('Hz').magnitude,
            np.abs(freqs_ts[nyquist_freq_idx].rescale('Hz').magnitude)) * pq.Hz

        np.testing.assert_array_equal(freqs_os, ts_overlap_freqs)

        np.testing.assert_allclose(
            cross_spec_os.magnitude,
            cross_spec_ts[:, :, ts_freq_indices].magnitude, rtol=1e-12, atol=0)

    def test_multitaper_cross_spectrum_parameter_hierarchy(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=(2, data_length))
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)
        signal_x = np.sin(2 * np.pi * signal_freq * time_points) + noise[0]
        signal_y = np.cos(2 * np.pi * signal_freq * time_points) + noise[1]
        data = AnalogSignal(np.vstack([signal_x, signal_y]).T,
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # Test num_tapers vs nw
        freqs1, cross_spec1 = elephant.spectral.multitaper_cross_spectrum(
            data, fs=data.sampling_rate, nw=3, num_tapers=9)
        freqs2, cross_spec2 = elephant.spectral.multitaper_cross_spectrum(
            data, fs=data.sampling_rate, nw=3)

        self.assertTrue((freqs1 == freqs2).all()
                        and (cross_spec1 != cross_spec2).all())

        # Test peak_resolution vs nw
        freqs1, cross_spec1 = elephant.spectral.multitaper_cross_spectrum(
            data, fs=data.sampling_rate, nw=3, num_tapers=9, peak_resolution=1)
        freqs2, cross_spec2 = elephant.spectral.multitaper_cross_spectrum(
            data, fs=data.sampling_rate, nw=3, num_tapers=9)

        self.assertTrue((freqs1 == freqs2).all()
                        and (cross_spec1 != cross_spec2).all())

    def test_multitaper_cross_spectrum_input_types(self):
        # generate a test data
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=(2, data_length))
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)
        signal_x = np.sin(2 * np.pi * signal_freq * time_points) + noise[0]
        signal_y = np.cos(2 * np.pi * signal_freq * time_points) + noise[1]
        data = AnalogSignal(np.vstack([signal_x, signal_y]).T,
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # outputs from AnalogSignal input are of Quantity type (standard usage)
        freqs_neo, cross_spec_neo \
            = elephant.spectral.multitaper_cross_spectrum(data)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(cross_spec_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, cross_spec_pq \
            = elephant.spectral.multitaper_cross_spectrum(
                    data.magnitude.T * data.units,
                    fs=1 / (sampling_period * pq.s))
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(cross_spec_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, cross_spec_np \
            = elephant.spectral.multitaper_cross_spectrum(
                    data.magnitude.T,
                    fs=1 / (sampling_period * pq.s))
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(cross_spec_np, pq.quantity.Quantity))

        # check if the results from different input types are identical
        self.assertTrue(
            (freqs_neo == freqs_pq).all() and
            (cross_spec_neo == cross_spec_pq).all())
        self.assertTrue(
            (freqs_neo == freqs_np).all() and
            (cross_spec_neo == cross_spec_np).all())


class SegmentedMultitaperCrossSpectrumTestCase(unittest.TestCase):
    def test_segmented_multitaper_cross_spectrum_errors(self):
        # generate dummy data
        data_length = 5000
        signal = AnalogSignal(np.zeros(data_length),
                              sampling_period=0.001 * pq.s,
                              units='mV')
        fs = signal.sampling_rate

        # - frequency resolution
        self.assertRaises(
            ValueError, elephant.spectral.segmented_multitaper_cross_spectrum,
            signal, fs=fs, frequency_resolution=-10)

        # - n per segment
        # n_per_seg = int(fs / dF), where dF is the frequency_resolution
        broken_freq_resolution = fs / (data_length+1)
        self.assertRaises(
            ValueError, elephant.spectral.segmented_multitaper_cross_spectrum,
            signal, fs=fs, frequency_resolution=broken_freq_resolution)

        # - length of segment (negative)
        self.assertRaises(
            ValueError, elephant.spectral.segmented_multitaper_cross_spectrum,
            signal, fs=fs, len_segment=-10)

        # - length of segment (larger than data length)
        self.assertRaises(
            ValueError, elephant.spectral.segmented_multitaper_cross_spectrum,
            signal, fs=fs, len_segment=data_length+1)

        # - number of segments (negative)
        self.assertRaises(
            ValueError, elephant.spectral.segmented_multitaper_cross_spectrum,
            signal, fs=fs, n_segments=-10)

        # - number of segments (larger than data length)
        self.assertRaises(
            ValueError, elephant.spectral.segmented_multitaper_cross_spectrum,
            signal, fs=fs, n_segments=data_length+1)

    def test_segmented_multitaper_cross_spectrum_behavior(self):
        # generate data (frequency domain to time domain)
        r = np.ones(2501) * 0.2
        r[0], r[500] = 0, 10  # Zero DC, peak at 100 Hz
        phi_x = np.random.uniform(-np.pi, np.pi, len(r))
        phi_y = np.random.uniform(-np.pi, np.pi, len(r))
        fake_coeffs_x = r*np.exp(1j * phi_x)
        fake_coeffs_y = r*np.exp(1j * phi_y)
        signal_x = scipy.fft.irfft(fake_coeffs_x)
        signal_y = scipy.fft.irfft(fake_coeffs_y)
        sampling_period = 0.001
        freqs = scipy.fft.rfftfreq(len(signal_x), d=sampling_period)
        signal_freq = freqs[r.argmax()]
        data = AnalogSignal(np.vstack([signal_x, signal_y]).T,
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # consistency between different ways of specifying n_per_seg
        # n_per_seg = int(fs/dF) and n_per_seg = len_segment
        frequency_resolution = 1 * pq.Hz
        len_segment = int(data.sampling_rate / frequency_resolution)

        freqs_fr, cross_spec_fr = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, frequency_resolution=frequency_resolution)

        freqs_ls, cross_spec_ls = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, len_segment=len_segment)

        np.testing.assert_array_equal(freqs_fr, freqs_ls)
        np.testing.assert_array_equal(cross_spec_fr, cross_spec_ls)

        # one-sided vs two-sided spectrum
        freqs_os, cross_spec_os = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, return_onesided=True)

        freqs_ts, cross_spec_ts = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, return_onesided=False)

        # test overlap parameter
        no_overlap = 0
        half_overlap = 0.5
        large_overlap = 0.99
        n_segments = 10

        freqs_no, cross_spec_no = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, n_segments=n_segments, overlap=no_overlap)

        freqs_ho, cross_spec_ho = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, n_segments=n_segments, overlap=half_overlap)

        freqs_lo, cross_spec_lo = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, n_segments=n_segments, overlap=large_overlap)

        self.assertTrue(freqs_no.shape < freqs_ho.shape < freqs_lo.shape)
        self.assertTrue(
            cross_spec_no.shape < cross_spec_ho.shape < cross_spec_lo.shape)

        # Nyquist frequency is negative when using onesided=False (fftfreq)
        # See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.fft.rfftfreq.html#scipy.fft.rfftfreq  # noqa
        nonnegative_freqs_indices = np.nonzero(freqs_ts >= 0)[0]
        nyquist_freq_idx = np.abs(freqs_ts).argmax()
        ts_freq_indices = np.append(nonnegative_freqs_indices,
                                    nyquist_freq_idx)
        ts_overlap_freqs = np.append(
            freqs_ts[nonnegative_freqs_indices].rescale('Hz').magnitude,
            np.abs(freqs_ts[nyquist_freq_idx].rescale('Hz').magnitude)) * pq.Hz

        np.testing.assert_array_equal(freqs_os, ts_overlap_freqs)

        np.testing.assert_allclose(
            cross_spec_os.magnitude,
            cross_spec_ts[:, :, ts_freq_indices].magnitude, rtol=1e-12, atol=0)

    def test_segmented_multitaper_cross_spectrum_parameter_hierarchy(self):
        # test frequency_resolution vs len_segment vs n_segments
        # generate data (frequency domain to time domain)
        r = np.ones(2501) * 0.2
        r[0], r[500] = 0, 10  # Zero DC, peak at 100 Hz
        phi_x = np.random.uniform(-np.pi, np.pi, len(r))
        phi_y = np.random.uniform(-np.pi, np.pi, len(r))
        fake_coeffs_x = r*np.exp(1j * phi_x)
        fake_coeffs_y = r*np.exp(1j * phi_y)
        signal_x = scipy.fft.irfft(fake_coeffs_x)
        signal_y = scipy.fft.irfft(fake_coeffs_y)
        sampling_period = 0.001
        freqs = scipy.fft.rfftfreq(len(signal_x), d=sampling_period)
        signal_freq = freqs[r.argmax()]
        data = AnalogSignal(np.vstack([signal_x, signal_y]).T,
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        n_segments = 5
        len_segment = 2000
        frequency_resolution = 1 * pq.Hz

        freqs_ns, cross_spec_ns = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, n_segments=n_segments)

        freqs_ls, cross_spec_ls = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, n_segments=n_segments, len_segment=len_segment)

        freqs_fr, cross_spec_fr = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, n_segments=n_segments, len_segment=len_segment,
                frequency_resolution=frequency_resolution)

        self.assertNotEqual(freqs_ns.shape, freqs_ls.shape)
        self.assertNotEqual(freqs_ls.shape, freqs_fr.shape)
        self.assertNotEqual(freqs_fr.shape, freqs_ns.shape)
        self.assertNotEqual(cross_spec_ns.shape, cross_spec_ls.shape)
        self.assertNotEqual(cross_spec_ls.shape, cross_spec_fr.shape)
        self.assertNotEqual(cross_spec_fr.shape, cross_spec_ns.shape)

    def test_segmented_multitaper_cross_spectrum_against_multitaper_psd(self):
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=(2, data_length))
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)
        signal_x = np.sin(2 * np.pi * signal_freq * time_points) + noise[0]
        signal_y = np.cos(2 * np.pi * signal_freq * time_points) + noise[1]
        data = AnalogSignal(np.vstack([signal_x, signal_y]).T,
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        freqs1, psd_multitaper = elephant.spectral.multitaper_psd(
            signal=data, fs=data.sampling_rate, nw=4, num_tapers=8)

        psd_multitaper[:, 1:] /= 2  # since comparing rfft and fft results

        freqs2, cross_spec = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data,
                fs=data.sampling_rate,
                nw=4,
                num_tapers=8,
                return_onesided=True)

        self.assertTrue((freqs1 == freqs2).all())

        np.testing.assert_allclose(psd_multitaper.magnitude,
                                   np.diagonal(cross_spec).T.real.magnitude,
                                   rtol=0.01,
                                   atol=0.01)

    def test_segmented_multitaper_cross_spectrum_input_types(self):
        # generate a test data
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=(2, data_length))
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)
        signal_x = np.sin(2 * np.pi * signal_freq * time_points) + noise[0]
        signal_y = np.cos(2 * np.pi * signal_freq * time_points) + noise[1]
        data = AnalogSignal(np.vstack([signal_x, signal_y]).T,
                            sampling_period=sampling_period * pq.s,
                            units='mV')

        # frequency resolution as an integer
        freq_res_int = 1
        freq_res_hz = 1 * pq.Hz

        freqs_int, cross_spec_int = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, frequency_resolution=freq_res_int)

        freqs_hz, cross_spec_hz = \
            elephant.spectral.segmented_multitaper_cross_spectrum(
                data, frequency_resolution=freq_res_hz)

        np.testing.assert_array_equal(freqs_int, freqs_hz)
        np.testing.assert_array_equal(cross_spec_int, cross_spec_hz)


class MultitaperCoherenceTestCase(unittest.TestCase):
    def test_multitaper_coherence_input_types(self):
        # Generate dummy data
        data_length = 10000
        sampling_period = 0.001
        signal_freq = 100.0
        np.random.seed(123)
        noise = np.random.normal(size=(2, data_length))
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)
        # Signals are designed to have coherence peak at `signal_freq`
        arr_signal_i = np.sin(2 * np.pi * signal_freq * time_points) + noise[0]
        arr_signal_j = np.cos(2 * np.pi * signal_freq * time_points) + noise[1]

        fs = 1000 * pq.Hz
        anasig_signal_i = neo.core.AnalogSignal(arr_signal_i,
                                                sampling_rate=fs,
                                                units=pq.mV)
        anasig_signal_j = neo.core.AnalogSignal(arr_signal_j,
                                                sampling_rate=fs,
                                                units=pq.mV)

        arr_f, arr_coh, arr_phi = elephant.spectral.multitaper_coherence(
            arr_signal_i, arr_signal_j, fs=fs)
        anasig_f, anasig_coh, anasig_phi = \
            elephant.spectral.multitaper_coherence(anasig_signal_i,
                                                   anasig_signal_j)

        np.testing.assert_array_equal(arr_f, anasig_f)
        np.testing.assert_allclose(arr_coh, anasig_coh, atol=1e-6)
        np.testing.assert_array_equal(arr_phi, anasig_phi)

    def test_multitaper_cohere_peak(self):
        # Generate dummy data
        data_length = 10000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=(2, data_length))
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)
        # Signals are designed to have coherence peak at `signal_freq`
        signal_i = np.sin(2 * np.pi * signal_freq * time_points) + noise[0]
        signal_j = np.cos(2 * np.pi * signal_freq * time_points) + noise[1]

        # Estimate coherence and phase lag with the multitaper method
        freq1, coh1, phase_lag1 = elephant.spectral.multitaper_coherence(
            signal_i,
            signal_j,
            fs=1/sampling_period,
            n_segments=16)

        indices, vals = scipy.signal.find_peaks(coh1, height=0.8, distance=10)

        peak_freqs = freq1[indices]

        np.testing.assert_allclose(peak_freqs,
                                   signal_freq*np.ones(len(peak_freqs)),
                                   rtol=0.05)

    @pytest.mark.skipif(version.parse(np.__version__)>version.parse("1.25.0"),
                        reason="This test will fail with numpy version"
                               "1.25.0 - 1.25.2,  see issue #24000"
                               "https://github.com/numpy/numpy/issues/24000 ")
    def test_multitaper_cohere_perfect_cohere(self):
        # Generate dummy data
        data_length = 10000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=(1, data_length))
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)
        signal = np.cos(2 * np.pi * signal_freq * time_points) + noise

        # Estimate coherence and phase lag with the multitaper method
        freq1, coh, phase_lag = elephant.spectral.multitaper_coherence(
            signal,
            signal,
            fs=1/sampling_period,
            n_segments=16)

        np.testing.assert_array_equal(phase_lag, np.zeros(phase_lag.size))
        np.testing.assert_array_equal(coh, np.ones(coh.size))

    def test_multitaper_cohere_no_cohere(self):
        # Generate dummy data
        data_length = 10000
        sampling_period = 0.001
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)

        signal_i = np.sin(2 * np.pi * 2.5 * time_points)
        signal_j = np.sin(2 * np.pi * 5 * time_points)

        # Estimate coherence and phase lag with the multitaper method
        freq, coh, phase_lag = elephant.spectral.multitaper_coherence(
            signal_i,
            signal_j,
            fs=1/sampling_period,
            n_segments=16)

        np.testing.assert_allclose(coh, np.zeros(coh.size), atol=0.002)

    def test_multitaper_cohere_phase_lag(self):
        # Generate dummy data
        data_length = 10000
        sampling_period = 0.001
        signal_freq = 100.0
        time_points = np.arange(0, data_length * sampling_period,
                                sampling_period)

        # Signals are designed to have maximal phase lag at 100 with value pi/4
        signal_i = np.sin(2 * np.pi * signal_freq * time_points + np.pi / 4)
        signal_j = np.cos(2 * np.pi * signal_freq * time_points)

        # Estimate coherence and phase lag with the multitaper method
        freq, coh, phase_lag = elephant.spectral.multitaper_coherence(
            signal_i,
            signal_j,
            fs=1/sampling_period,
            n_segments=16,
            num_tapers=8)

        indices, vals = scipy.signal.find_peaks(phase_lag,
                                                height=0.8 * np.pi / 4,
                                                distance=10)

        # Get peak frequencies and peak heights
        peak_freqs = freq[indices]
        peak_heights = vals['peak_heights']

        np.testing.assert_allclose(peak_freqs,
                                   signal_freq*np.ones(len(peak_freqs)),
                                   rtol=0.05)
        np.testing.assert_allclose(peak_heights,
                                   np.pi / 4 * np.ones(len(peak_heights)),
                                   rtol=0.05)


class WelchCohereTestCase(unittest.TestCase):
    def test_welch_cohere_errors(self):
        # generate a dummy data
        x = AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                         units='mV')
        y = AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                         units='mV')

        # check for invalid parameter values
        # - length of segments
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          len_segment=0)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          len_segment=x.shape[0] * 2)
        # - number of segments
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          n_segments=0)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          n_segments=x.shape[0] * 2)
        # - frequency resolution
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          frequency_resolution=-1)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          frequency_resolution=x.sampling_rate /
                          (x.shape[0] + 1))
        # - overlap
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          overlap=-1.0)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          overlap=1.1)

    def test_welch_cohere_warnings(self):
        # generate a dummy data
        x = AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                         units='mV')
        y = AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                         units='mV')
        # Test deprecation warning for 'hanning' window
        self.assertWarns(DeprecationWarning, elephant.spectral.welch_coherence,
                         x, y, window='hanning')

    def test_welch_cohere_behavior(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise1 = np.random.normal(size=data_length) * 0.01
        noise2 = np.random.normal(size=data_length) * 0.01
        signal1 = [np.cos(2 * np.pi * signal_freq * t)
                   for t in np.arange(0, data_length * sampling_period,
                                      sampling_period)]
        signal2 = [np.sin(2 * np.pi * signal_freq * t)
                   for t in np.arange(0, data_length * sampling_period,
                                      sampling_period)]
        x = AnalogSignal(np.array(signal1 + noise1), units='mV',
                         sampling_period=sampling_period * pq.s)
        y = AnalogSignal(np.array(signal2 + noise2), units='mV',
                         sampling_period=sampling_period * pq.s)

        # consistency between different ways of specifying segment length
        freqs1, coherency1, phase_lag1 = elephant.spectral.welch_coherence(
            x, y, len_segment=data_length // 5, overlap=0)
        freqs2, coherency2, phase_lag2 = elephant.spectral.welch_coherence(
            x, y, n_segments=5, overlap=0)
        self.assertTrue((coherency1 == coherency2).all() and
                        (phase_lag1 == phase_lag2).all() and
                        (freqs1 == freqs2).all())

        # frequency resolution and consistency with data
        freq_res = 1.0 * pq.Hz
        freqs, coherency, phase_lag = elephant.spectral.welch_coherence(
            x, y, frequency_resolution=freq_res)
        self.assertAlmostEqual(freq_res, freqs[1] - freqs[0])
        self.assertAlmostEqual(freqs[coherency.argmax()], signal_freq,
                               places=2)
        self.assertAlmostEqual(phase_lag[coherency.argmax()], -np.pi / 2,
                               places=2)
        freqs_np, coherency_np, phase_lag_np = \
            elephant.spectral.welch_coherence(x.magnitude.flatten(),
                                              y.magnitude.flatten(),
                                              fs=1 / sampling_period,
                                              frequency_resolution=freq_res)
        assert_array_equal(freqs.simplified.magnitude, freqs_np)
        assert_array_equal(coherency[:, 0], coherency_np)
        assert_array_equal(phase_lag[:, 0], phase_lag_np)

        # - check the behavior of parameter `axis` using multidimensional data
        num_channel = 4
        data_length = 5000
        x_multidim = np.random.normal(size=(num_channel, data_length))
        y_multidim = np.random.normal(size=(num_channel, data_length))
        freqs, coherency, phase_lag = \
            elephant.spectral.welch_coherence(x_multidim, y_multidim)
        freqs_T, coherency_T, phase_lag_T = elephant.spectral.welch_coherence(
            x_multidim.T, y_multidim.T, axis=0)
        assert_array_equal(freqs, freqs_T)
        assert_array_equal(coherency, coherency_T.T)
        assert_array_equal(phase_lag, phase_lag_T.T)

    def test_welch_cohere_input_types(self):
        # generate a test data
        sampling_period = 0.001
        x = AnalogSignal(np.array(np.random.normal(size=5000)),
                         sampling_period=sampling_period * pq.s,
                         units='mV')
        y = AnalogSignal(np.array(np.random.normal(size=5000)),
                         sampling_period=sampling_period * pq.s,
                         units='mV')

        # outputs from AnalogSignal input are of Quantity type
        # (standard usage)
        freqs_neo, coherency_neo, phase_lag_neo = \
            elephant.spectral.welch_coherence(x, y)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(phase_lag_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, coherency_pq, phase_lag_pq = elephant.spectral \
            .welch_coherence(x.magnitude.flatten() * x.units,
                             y.magnitude.flatten() * y.units,
                             fs=1 / sampling_period)
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(phase_lag_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, coherency_np, phase_lag_np = elephant.spectral \
            .welch_coherence(x.magnitude.flatten(),
                             y.magnitude.flatten(),
                             fs=1 / sampling_period)
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(phase_lag_np, pq.quantity.Quantity))

        # check if the results from different input types are identical
        self.assertTrue((freqs_neo == freqs_pq).all() and
                        (coherency_neo[:, 0] == coherency_pq).all() and
                        (phase_lag_neo[:, 0] == phase_lag_pq).all())
        self.assertTrue((freqs_neo == freqs_np).all() and
                        (coherency_neo[:, 0] == coherency_np).all() and
                        (phase_lag_neo[:, 0] == phase_lag_np).all())

    def test_welch_cohere_multidim_input(self):
        # generate multidimensional data
        num_channel = 4
        data_length = 5000
        sampling_period = 0.001
        x_np = np.array(np.random.normal(size=(num_channel, data_length)))
        y_np = np.array(np.random.normal(size=(num_channel, data_length)))
        # Since row-column order in AnalogSignal is different from the
        # convention in NumPy/SciPy, `data_np` needs to be transposed when it's
        # used to define an AnalogSignal
        x_neo = AnalogSignal(x_np.T, units='mV',
                             sampling_period=sampling_period * pq.s)
        y_neo = AnalogSignal(y_np.T, units='mV',
                             sampling_period=sampling_period * pq.s)
        x_neo_1dim = AnalogSignal(x_np[0], units='mV',
                                  sampling_period=sampling_period * pq.s)
        y_neo_1dim = AnalogSignal(y_np[0], units='mV',
                                  sampling_period=sampling_period * pq.s)

        # check if the results from different input types are identical
        freqs_np, coherency_np, phase_lag_np = elephant.spectral \
            .welch_coherence(x_np, y_np, fs=1 / sampling_period)
        freqs_neo, coherency_neo, phase_lag_neo = \
            elephant.spectral.welch_coherence(x_neo, y_neo)
        freqs_neo_1dim, coherency_neo_1dim, phase_lag_neo_1dim = \
            elephant.spectral.welch_coherence(x_neo_1dim, y_neo_1dim)
        self.assertTrue(np.all(freqs_np == freqs_neo))
        self.assertTrue(np.all(coherency_np.T == coherency_neo))
        self.assertTrue(np.all(phase_lag_np.T == phase_lag_neo))
        self.assertTrue(
            np.all(coherency_neo_1dim[:, 0] == coherency_neo[:, 0]))
        self.assertTrue(
            np.all(phase_lag_neo_1dim[:, 0] == phase_lag_neo[:, 0]))


if __name__ == "__main__":
    unittest.main(verbosity=2)



import itertools
from typing import Iterable, List, NamedTuple, Union, Optional

import numpy as np
from scipy.signal import oaconvolve

from elephant.conversion import BinnedSpikeTrain


def total_spiking_probability_edges(
    spike_trains: BinnedSpikeTrain,
    surrounding_window_sizes: Optional[List[int]] = None,
    observed_window_sizes: Optional[List[int]] = None,
    crossover_window_sizes: Optional[List[int]] = None,
    max_delay: int = 25,
    normalize: bool = False,
):
    r"""
    Estimate the functional connectivity and delay times of a neural network
    using Total Spiking Probability Edges (TSPE).

    This algorithm uses a normalized cross correlation between pairs of
    spike trains at different delay times to get a cross-correlogram.
    Afterwards, a series of convolutions with multiple edge filters
    on the cross-correlogram are performed, in order to estimate the
    connectivity between neurons and thus allowing the discrimination
    between inhibitory and excitatory effects.

    The default window sizes and maximum delay were optimized using
    in-silico generated spike trains.

    **Background:**

    - On an excitatory connection the spike rate increases and decreases again
      due to the refractory period which results in local maxima in the
      cross-correlogram followed by downwards slope.
    - On an inhibitory connection the spike rate decreases and after refractory
      period, increases again which results in local minima surrounded by high
      values in the cross-correlogram.
    - An edge filter can be used to interpret the cross-correlogram and
      accentuate the local maxima and minima

    **Procedure:**

    1. Compute normalized cross-correlation :math:`NCC` of spike trains of all
       neuron pairs.
    2. Convolve :math:`NCC` with edge filter :math:`g_{i}` to compute
       :math:`SPE`.
    3. Convolve :math:`SPE` with corresponding running total filter
       :math:`h_{i}` to account for different lengths after convolution with
       edge filter.
    4. Compute :math:`TSPE` using the sum of all :math:`SPE` for all different
       filter pairs.
    5. Compute the connectivity matrix by using the index of the TSPE values
       with the highest absolute values.

    **Normalized Cross-Correlation:**

    .. math::

        NCC_{XY}(d) = \frac{1}{N} \sum_{i=-\infty}^{\infty}{ \frac{ (y_{(i)} -
        \bar{y}) \cdot (x_{(i-d)} - \bar{x}) }{ \sigma_x \cdot \sigma_y }}

    **Edge Filter**

    .. math::

        g_{(i)} = \begin{cases}
        - \frac{1}{a} & 0 \lt i \leq a \ \
        \frac{2}{b} & a+c \lt i \leq a + b + c \ \
        - \frac{1}{a} & a+b+2c \lt i \leq 2a + b + 2c \ \
        0 & \mathrm{otherwise}
        \end{cases}

    where :math:`a` is the parameter `surrounding_window_size`, :math:`b`
    `observed_window_size`, and :math:`c` is the parameter
    `crossover_window_size`.


**Spiking Probability Edges**

.. math::
    SPE_{X \rightarrow Y(d)} = NCC_{XY}(d) * g(i)

*Total Spiking Probability Edges:*

.. math::
    TSPE_{X \rightarrow Y}(d) = \sum_{n=1}^{N_a \cdot N_b \cdot N_c}
    {SPE_{X \rightarrow Y}^{(n)}(d) * h(i)^{(n)} }

:cite:`functional_connectivity-de_blasi19_169`

Parameters
----------
spike_trains : (N, ) elephant.conversion.BinnedSpikeTrain
    A binned spike train containing all neurons for connectivity estimation
surrounding_window_sizes : List[int]
    Array of window sizes for the surrounding area of the point of
    interest.  This corresponds to parameter `a` of the edge filter in
    :cite:`functional_connectivity-de_blasi19_169`. Value is given in units of
    the number of bins according to the binned spike trains `spike_trains`.
    Default: [3, 4, 5, 6, 7, 8]
observed_window_sizes : List[int]
    Array of window sizes for the observed area. This corresponds to
    parameter `b` of the edge filter and the length of the running filter
    as defined in :cite:`functional_connectivity-de_blasi19_169`. Value is
    given in units of the number of bins according to the binned spike trains
    `spike_trains`.
    Default: [2, 3, 4, 5, 6]
crossover_window_sizes : List[int]
    Array of window sizes for the crossover between surrounding and
    observed window. This corresponds to parameter `c` of the edge filter in
    :cite:`functional_connectivity-de_blasi19_169`. Value is given in units of
    the number of bins according to the binned spike trains `spike_trains`.
    Default: [0]
max_delay : int
    Defines the max delay when performing the normalized cross-correlations.
    Value is given in units of the number of bins according to the binned spike
    trains `spike_trains`.
    Default: 25
normalize : bool, optional
    Normalize the output [experimental]. Default: False.

Returns
-------
connectivity_matrix : (N, N) np.ndarray
    Square matrix of the connectivity estimation between neurons.
    Positive values describe an excitatory connection while
    negative values describe an inhibitory connection.
delay_matrix : (N, N) np.ndarray
    Square matrix of the estimated delay times between neuron activities.
"""

    if not surrounding_window_sizes:
        surrounding_window_sizes = [3, 4, 5, 6, 7, 8]

    if not observed_window_sizes:
        observed_window_sizes = [2, 3, 4, 5, 6]

    if not crossover_window_sizes:
        crossover_window_sizes = [0]

    n_neurons, n_bins = spike_trains.shape

    filter_pairs = generate_filter_pairs(
        surrounding_window_sizes, observed_window_sizes, crossover_window_sizes
    )

    # Calculate normalized cross-correlation for different delays.
    # The delay range is from 0 to max_delay and includes
    # padding for the filter convolution
    max_padding = max(surrounding_window_sizes) + max(crossover_window_sizes)
    delay_times = list(range(-max_padding, max_delay + max_padding))
    NCC_d = normalized_cross_correlation(spike_trains, delay_times=delay_times)

    # Normalize to counter network bursts
    if normalize:
        for delay_time in delay_times:
            NCC_d[:, :, delay_time] /= np.sum(
                NCC_d[:, :, delay_time][~np.identity(NCC_d.shape[0],
                                                     dtype=bool)]
            )

    # Apply edge and running total filter
    tspe_matrix = np.zeros((n_neurons, n_neurons, max_delay))
    for filter in filter_pairs:
        # Select ncc_window based on needed filter padding
        NCC_window = NCC_d[
            :,
            :,
            max_padding
            - filter.needed_padding: max_delay
            + max_padding
            + filter.needed_padding,
        ]

        # Compute two convolutions with edge- and running total filter
        x1 = oaconvolve(
            NCC_window, np.expand_dims(filter.edge_filter, (0, 1)),
            mode="valid", axes=2
        )
        x2 = oaconvolve(
            x1, np.expand_dims(filter.running_total_filter, (0, 1)),
            mode="full", axes=2
        )

        tspe_matrix += x2

    # Take maxima of absolute of delays to get estimation for connectivity
    connectivity_matrix_index = np.argmax(np.abs(tspe_matrix),
                                          axis=2, keepdims=True)
    connectivity_matrix = np.take_along_axis(tspe_matrix,
                                             connectivity_matrix_index, axis=2
                                             ).squeeze(axis=2)
    delay_matrix = connectivity_matrix_index.squeeze()

    return connectivity_matrix, delay_matrix


def normalized_cross_correlation(
    spike_trains: BinnedSpikeTrain,
    delay_times: Union[int, List[int], Iterable[int]] = 0,
) -> np.ndarray:
    r"""
    Normalized cross correlation using std deviation

    Computes the normalized_cross_correlation between all
    spike trains inside a `BinnedSpikeTrain` object at a given delay time.

    The underlying formula is:

    .. math::
        NCC_{X\arrY(d)} = \frac{1}{N_{bins}}\sum_{i=-\inf}^{\inf}{
        \frac{(y_{(i)} - \bar{y}) \cdot (x_{(i-d) - \bar{x})}{\sigma_x
        \cdot \sigma_y}}}

    The subtraction of mean-values is omitted, since it offers little added
    accuracy but increases the compute-time considerably.
    """

    n_neurons, n_bins = spike_trains.shape

    # Get sparse array of BinnedSpikeTrain
    spike_trains_array = spike_trains.sparse_matrix

    # Get std deviation of spike trains
    spike_trains_zeroed = spike_trains_array - spike_trains_array.mean(axis=1)
    spike_trains_std = np.std(spike_trains_zeroed, ddof=1, axis=1)
    std_factors = spike_trains_std @ spike_trains_std.T

    # Loop over delay times
    if isinstance(delay_times, int):
        delay_times = [delay_times]
    elif isinstance(delay_times, list):
        pass
    elif isinstance(delay_times, Iterable):
        delay_times = list(delay_times)

    NCC_d = np.zeros((len(delay_times), n_neurons, n_neurons))

    for index, delay_time in enumerate(delay_times):
        # Uses theoretical zero-padding for shifted values,
        # but since $0 \cdot x = 0$ values can simply be omitted
        if delay_time == 0:
            CC = spike_trains_array[:, :] @ spike_trains_array[:, :
                                                               ].transpose()

        elif delay_time > 0:
            CC = (
                spike_trains_array[:, delay_time:]
                @ spike_trains_array[:, :-delay_time].transpose()
            )

        else:
            CC = (
                spike_trains_array[:, :delay_time]
                @ spike_trains_array[:, -delay_time:].transpose()
            )

        # Convert CC to dense matrix before performing the division
        CC = CC.toarray()
        # Normalize using std deviation
        NCC = CC / std_factors / n_bins

        # Compute cross correlation at given delay time
        NCC_d[index, :, :] = NCC

    # Move delay_time axis to back of array
    # Makes index using neurons more intuitive → (n_neuron, n_neuron,
    #                                             delay_times)
    NCC_d = np.moveaxis(NCC_d, 0, -1)

    return NCC_d


def generate_edge_filter(
    surrounding_window_size: int,
    observed_window_size: int,
    crossover_window_size: int,
) -> np.ndarray:
    r"""Generate an edge filter

    The edge filter is generated using following piecewise defined function:

    a = surrounding_window_size
    b = observed_window_size
    c = crossover_window_size

    .. math::
        g_{(i)} = \begin{cases}
            - \frac{1}{a} & 0 \lt i \leq a \\
            \frac{2}{b} & a+c \lt i \leq a + b + c \\
            - \frac{1}{a} & a+b+2c \lt i \leq 2a + b + 2c \ \
            0 & \mathrm{otherwise}
            \end{cases}

    """
    filter_length = (
        (2 * surrounding_window_size)
        + observed_window_size
        + (2 * crossover_window_size)
    )
    i = np.arange(1, filter_length + 1, dtype=np.float64)

    conditions = [
        (i > 0) & (i <= surrounding_window_size),
        (i > (surrounding_window_size + crossover_window_size))
        & (i <= surrounding_window_size + observed_window_size +
           crossover_window_size),
        (
            i
            > surrounding_window_size
            + observed_window_size
            + (2 * crossover_window_size)
        )
        & (
            i
            <= (2 * surrounding_window_size)
            + observed_window_size
            + (2 * crossover_window_size)
        ),
    ]

    values = [
        -(1 / surrounding_window_size),
        2 / observed_window_size,
        -(1 / surrounding_window_size),
        0,
    ]  # Default Value

    edge_filter = np.piecewise(i, conditions, values)

    return edge_filter


def generate_running_total_filter(observed_window_size: int) -> np.ndarray:
    return np.ones(observed_window_size)


class TspeFilterPair(NamedTuple):
    edge_filter: np.ndarray
    running_total_filter: np.ndarray
    needed_padding: int
    surrounding_window_size: int
    observed_window_size: int
    crossover_window_size: int


def generate_filter_pairs(
    surrounding_window_sizes: List[int],
    observed_window_sizes: List[int],
    crossover_window_sizes: List[int],
) -> List[TspeFilterPair]:
    """Generates filter pairs of edge and running total filter using all
    permutations of given parameters
    """
    filter_pairs = []

    for _a, _b, _c in itertools.product(
        surrounding_window_sizes, observed_window_sizes, crossover_window_sizes
    ):
        edge_filter = generate_edge_filter(_a, _b, _c)
        running_total_filter = generate_running_total_filter(_b)
        needed_padding = _a + _c
        filter_pair = TspeFilterPair(
            edge_filter, running_total_filter, needed_padding, _a, _b, _c
        )
        filter_pairs.append(filter_pair)

    return filter_pairs
